from ._example_1 import example as example_1
from ._example_2 import example as example_2
from ._example_3 import example as example_3
from ._example_4 import example as example_4
from ._example_5 import example as example_5

# Add more examples as needed
from ._types import ICLExample

icl_examples: list[ICLExample] = [example_1, example_2, example_3, example_4, example_5]
