# SPDX-License-Identifier: Apache-2.0
"""Shared data structures and functions for formatting code.

Data structures and functions that are shared across various types of input
and output formatting code.
"""
