from ._example_1 import example as example_1
from ._example_2 import example as example_2
from ._example_3 import example as example_3
from ._types import ICLExample

icl_examples: list[ICLExample] = [example_1, example_2, example_3]
