Think about what to do next and take the next step in the process. If a tool call is necessary to collect additional information call it. If you are done and have a final answer, call the `final_answer` tool.
